/*
 * UNC0xx host-side software. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc0xx.c
 * Set of common functions for communicating with the device
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "unc_data.h"
#include "unc_prot.h"

#include "unc0xx.h"
#include "unc0xx_hdl.h"

static int send_command(struct unc0xx_devhdl *dev_hdl,
                        int cmd, int param, int p2, uint32_t data)
{
    struct {
        uint8_t req_num, cmd, param;
        uint8_t param2;
        uint32_t data;
    } buf;
    int r;

    buf.req_num = uncreq_command;
    buf.cmd = cmd;
    buf.param = param;
    buf.param2 = p2;
    buf.data = data;
    r = dev_hdl->vtbl->send(dev_hdl,  &buf, sizeof(buf));
    if(r != sizeof(buf)) {
        unc0xx_set_errormsg(dev_hdl,
            "error sending info to device (wrong count)",
            unc0xx_msg_const);
        return -1;
    }
    return 0;
}

static int unc0xx_get_structure(struct unc0xx_devhdl *dev_hdl,
                                int req, void *data, int sz)
{
    int r;
    r = dev_hdl->vtbl->get(dev_hdl, req, data, sz);
    if(r <= 0) {
        return -1;
    }
    if(r != sz) {
        unc0xx_set_errormsg(dev_hdl,
            "wrong count on reading from the device",
            unc0xx_msg_const);
        return -1;
    }
    return 0;
}

int unc0xx_get_version(struct unc0xx_devhdl *dev_hdl)
{
    return dev_hdl->version;
}

int unc0xx_get_fullstate(struct unc0xx_devhdl *dev_hdl,
                         struct unc0xx_global_state_str *data)
{
    return unc0xx_get_structure(dev_hdl, uncreq_query_fullstate,
                                data, sizeof(*data));
}

int unc0xx_get_extrastate(struct unc0xx_devhdl *dev_hdl,
                          struct unc0xx_extra_state_str_common *data)
{
    return unc0xx_get_structure(dev_hdl, uncreq_query_extrastate,
                                data, sizeof(*data));
}

int unc0xx_get_configuration(struct unc0xx_devhdl *dev_hdl,
                             struct unc0xx_global_configuration_str *data)
{
    int rc = unc0xx_get_structure(dev_hdl, uncreq_query_configuration,
                                  data, sizeof(*data));
    if(rc > 0) {
        dev_hdl->version = data->version;
    }
    return rc;
}

int unc0xx_save_to_eeprom(struct unc0xx_devhdl *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_save_to_eeprom, 0, 0, 0);
}
int unc0xx_restore_from_eeprom(struct unc0xx_devhdl *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_restore_from_eeprom, 0, 0, 0);
}
int unc0xx_restore_factory(struct unc0xx_devhdl *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_restore_factory, 0, 0, 0);
}
int unc0xx_clear_msg(struct unc0xx_devhdl *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_clear_msg, 0, 0, 0);
}
int unc0xx_reset_1wire(struct unc0xx_devhdl *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_reset_1wire, 0, 0, 0);
}

int unc0xx_set_new_id(struct unc0xx_devhdl *dev_hdl, unsigned int id)
{
    return send_command(dev_hdl, uncreqcmd_set_id, 0, 0, id);
}

int unc0xx_set_configuration(struct unc0xx_devhdl *dev_hdl, unsigned int id)
{
    return send_command(dev_hdl, uncreqcmd_set_userconfig, 0, 0, id);
}

/* use with EXTREME CAUTION!!! better don't use at all */
int unc0xx_set_hardware_conf(struct unc0xx_devhdl *dev_hdl, unsigned int id)
{
    return send_command(dev_hdl, uncreqcmd_set_hwconfig, 0, 0, id);
}

int unc0xx_perform_action(struct unc0xx_devhdl *dev_hdl, int action)
{
    return send_command(dev_hdl, uncreqcmd_action, 0, 0, action);
}

int unc0xx_delayed_action(struct unc0xx_devhdl *dev_hdl,
                          int action, int delay, int location)
{
#if 0
    return send_command(dev_hdl, uncreqcmd_delayed_action,
                        act & 0xff, (act >> 8) & 0xff, dely);
#endif
    struct longcmd {  /* must be 16 bytes long */
        uint8_t req, cmd;
        uint16_t action;
        uint8_t location;
        uint8_t reserved[3];
        uint32_t delay;
        uint8_t reserved2[4];
    } buf;
    int k;
    buf.req = uncreq_long_command;
    buf.cmd = uncreqlcmd_delayed_action;
    buf.action = action;
    buf.location = location;
    buf.delay = delay;
    buf.reserved[0] = buf.reserved[1] = buf.reserved[2] = 0;
    k = dev_hdl->vtbl->send(dev_hdl, &buf, sizeof(buf));
    return k==16 ? 0 : -1;
}

#if 0
int unc0xx_delayed_action(struct unc0xx_devhdl *dev_hdl,
                          int time, int units, int action)
{
    uint32_t t, u, a;
    t = time & 0xff;
    t <<= 24;
    u = units & 0xff;
    u <<= 16;
    a = action & 0xffff;
    return send_command(dev_hdl, uncreqcmd_delayed_action, 0, 0, t | u | a);
}
#endif

int unc0xx_set_reactions(struct unc0xx_devhdl *dev_hdl,
                         int line, unsigned long reaction)
{
    return send_command(dev_hdl, uncreqcmd_set_reactions, line, 0, reaction);
}

int unc0xx_get_counter_range(struct unc0xx_devhdl *dev_hdl, int *min, int *max)
{
    struct unc0xx_global_configuration_str conf;
    int r = unc0xx_get_configuration(dev_hdl, &conf);
    if(r>=0) {
        *min = ((conf.hardware_config & 0x0f) == 0x04) ? 0 : 4;
        *max = ((conf.hardware_config & 0xf0) == 0x40) ? 7 : 3;
    }
    return r;
}

int unc0xx_zero_query_counters(struct unc0xx_devhdl *dev_hdl,
                               int counters[], int ncount)
{
    uint8_t data[1+2*UNC_INPUTS_COUNT];
    int r, i, min=0, max=-1;
    unc0xx_get_counter_range(dev_hdl, &min, &max);
    r = dev_hdl->vtbl->get(dev_hdl,
                           uncreq_zero_query_counters,
                           data, sizeof(data));
    if(r != sizeof(data))
        return -1;
    for(i=0; i<ncount; i++) {
        int i2 = i + min;
        if(i2 > max) break;
        counters[i] = data[2*i2+1] | ((uint16_t)(data[2*i2+2])<<8) ;
    }
    return i;
}

int unc0xx_dump_eeprom(struct unc0xx_devhdl *dev_hdl, void* buf)
{
    int sz = dev_hdl->version > 4000000 ? 64 : 128;
    return dev_hdl->vtbl->get(dev_hdl, uncreq_query_eeprom, buf, sz);
}

int unc0xx_takeover_stream(struct unc0xx_devhdl *dev_hdl)
{
    return dev_hdl->vtbl->takeover_stream(dev_hdl);
}

int unc0xx_read_stream(struct unc0xx_devhdl *dev_hdl, char buf[8])
{
    return dev_hdl->vtbl->read_stream(dev_hdl, buf);
}

int unc0xx_write_e_reg(struct unc0xx_devhdl *dev_hdl, int regn, char data[8])
{
    char buf[16]; // must be 16 as the request size is 16
    int k;
    int i;
    buf[0] = uncreq_long_command;
    buf[1] = uncreqlcmd_write_e_reg;
    buf[2] = regn;
    for(i = 0; i < 8; i++)
        buf[i+3] = data[i]; 
    k = dev_hdl->vtbl->send(dev_hdl, buf, sizeof(buf));
    return k==16 ? 0 : -1;
}

void unc0xx_close(struct unc0xx_devhdl *dev_hdl)
{
    dev_hdl->vtbl->close(dev_hdl);
}

/*
 * These functions are not static but intentionally not included into 
 * the header because they are NOT intended for normal use; the only
 * program which uses them is wrongreq.c; don't use these functions
 * unless you're absolutely sure you know what you're doing.
 */
int unc0xx_simple_request(struct unc0xx_devhdl *dev_hdl, int req)
{
    return send_command(dev_hdl, req, 0, 0, 0);
}
int unc0xx_raw_data_request(struct unc0xx_devhdl *dev_hdl,
                            int query_id, void *buf, int len)
{
    return dev_hdl->vtbl->get(dev_hdl, query_id, buf, len);
}

/* internal non-virtual methods */
void unc0xx_set_errormsg(struct unc0xx_devhdl *h, const char *msg, int mode)
{
    if(h->message && h->message_owned) {
        free((char*)(h->message));
        h->message = NULL;
    }
    if(!msg)
        return;
    switch(mode) {
        case unc0xx_msg_const:
            h->message = msg;
            h->message_owned = 0;
            break;
        case unc0xx_msg_own:
            h->message = msg;
            h->message_owned = 1;
            break;
        case unc0xx_msg_copy:
            h->message = strdup(msg);
            h->message_owned = 1;
            break;
    }
}

const char *unc0xx_get_errormsg(struct unc0xx_devhdl *h)
{
    return h->message;
}
