/*
 * UNC0xx host-side software. (c) Unicontrollers, Ltd., 2010-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file uncctl.c
 * The main file of the command line tool named uncctl, which
 * allows to control the UNC0xx device in batch mode
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdio.h>
#ifdef WIN32
#include <getopt.h>
#include <conio.h>
#else
#include <unistd.h>
#endif
#include <string.h>
#include <stdlib.h>

#include <stdint.h>


#include "unc0xx.h"
#include "unc_prot.h"
#include "unc_data.h"

static int hexdig2num(char c)
{
    if(c >= '0' && c <= '9') return c - '0';    
    if(c >= 'A' && c <= 'F') return c - 'A' + 10;    
    if(c >= 'a' && c <= 'f') return c - 'a' + 10;    
    return -1;
}

static void help()
{
    static const char *htext[] = {
        "usage: uncctl <opmode> <options>",
        "  opmodes: -L            find and list UNC0xx devices",
        "           -Q            dump the full state & configuration",
        "           -Z            query and zero input counters",
        "           -D            dump eeprom (for debug only)",
        "           -T <new id>   set UNC ID for the device",
        "           -G <conf_reg> change configuration register",
        "           -H <hw_conf> \"really\"   change HW CONF (dangerous!)",
        "           -A <action>   perform \"action\"",
        "           -R <in> <val> set reaction register for the <in> (0..7)",
        "                      to <val> (hex number or 'num:action:action')",
        "           -W            write configuration to eeprom",
        "           -E            restore configuration from eeprom",
        "           -F            restore factory defaults",
        "           -Y <nn> <hex_data> set Enn register's contents", 
        "                      <hex_data> must be exactly 16 hex-digits",
        "           -I            display the input stream from the device",
#ifdef WIN32
        "                         DOESN'T WORK under Win32 systems",
#endif
        "           -K            clear message buffer in the glob. state",
        "",
        "  options: -i <id>       use the device with this id",
        "                 (otherwise, the first found one is used)",
        "           -s <port>     use rs485 on the <port>, instead of USB",
        "           -b <baud>     speed (baud rate) for rs485:"
        "                 4800, 9600, 19200 (default), 38400, 57600"
        "           -t <msec>     set timeout (ms), default 5000",
        "           -d <delay>    with action, perform it after the delay",
        "                 e.g. '2h' is 2 hours, '7m' is 7 minutes,",
        "                 '15s' is 15 seconds, 120c or 120 is 120 cycles,",
        "           -l <location> with delayed action, places it to the",
        "                 given table row; <location> must be one digit,",
        "                 0, 1, ... 9", 
#ifndef WIN32
        "           -w            suppress the non-root warning",
#endif
        "           -v            be verbose (useful with -Q)",
        "           -h            display this help screen",
        "",
        "NOTE: Device ids must be given as hex word, e.g. 21af.",
        "      Action may be specified as hex (with leading `x'),",
        "      binary or four action chars (+,-,^,=), where 01 or +",
        "      means turn on, 10 or - means turn off, 11 or ^ means",
        "      switch to the opposit state and 00 or = means no action,",
        "      so \"x63\", \"01100011\" and \"+-=^\" are the same:",
        "      lineD on, lineC off, leave lineB as is, switch lineA",
        "   some examples of actions:",
        "      xAA    10101010    ----     all lines off",
        "      x55    01010101    ++++     all lines on",
        "      xff    11111111    ^^^^     switch all lines",
        "      x00    00000000    ====     no operation",
        "      x04    00000100    ==+=     lineB on, don't touch others",
        "  You can use actions with -R as well, e.g., \"-R 3 6000:++++:----\"",
        "  is the same as \"-R 3 600055AA\"",
        "",
        "  For the x9xx actions, you can also use @-mnemonics, such as",
        "  ++@26, +-@x1a, --@17, etc., where ++/-- are the bits to be sent,",
        "  and the number after '@' (either hex or decimal) is the E-reg",
        "  where the 1-Wire id is stored.",
        NULL
    };
    int i;
    printf("uncctl vers. 4.5 [%d] (c) Unicontrollers Ltd., 2010-2014\n",
           UNC_VERSION);
    for(i=0; htext[i]; i++)
        printf("%s\n", htext[i]);
}

/* command line parameters */

static enum {
    opm_undef,
    opm_list,
    opm_query,
    opm_zero,
    opm_dump_eeprom,
    opm_setid,
    opm_set_conf,
    opm_set_hwconf,
    opm_write_eeprom,
    opm_restore_from_eeprom,
    opm_restore_factory,
    opm_write_e_reg,
    opm_clear_messagebuf,
    opm_input_stream,
    opm_reset_1wire,
    opm_action,
#if 0
    opm_message,
    opm_intercept_com,
#endif
    opm_set_reactions
} operation_mode = opm_undef;
int device_id = 0;
int be_verbose = 0;
char **rest_args;
char *action_arg;
uint32_t action_delay = 0;
int action_loc = 0;
int use_rs485 = 0;
char *rs485_port = NULL;
int rs485_speed = 19200;
int warn_nonroot = 1;

static int set_op_mode(int m)
{
    if(operation_mode != opm_undef) {
        fprintf(stderr, "You can only specify ONE operation mode\n");
        return 0;
    }
    operation_mode = m;
    return 1;
}

static int extract_delay_value(const char *str)
{
    int dval = 0;
    while(*str >= '0' && *str <= '9') {
        dval *= 10;
        dval += (*str - '0');
        str++;
    }
    if(!dval) {
        fprintf(stderr, "Illegal delay value\n");
        return 0;
    }
    switch(*str) {
    case 0:
    case 'c':
    case 'C':
        break;
    case 's':
    case 'S':
        dval *= 100; break;
    case 'm':
    case 'M':
        dval *= (60 * 100); break;
    case 'h':
    case 'H':
        dval *= (60 * 60 * 100); break;
    default:
        fprintf(stderr, "Unknown units for the delay value\n");
        return 1;
    }
    action_delay = dval;
    return 1;
}

static int extract_baudrate(const char *rate)
{
    long res;
    char *err;
    res = strtol(rate, &err, 10);
    if(*err)
        return 0;
    switch(res) {
        case   4800:
        case   9600:
        case  19200:
        case  38400:
        case  57600:
        /* case 115200: */
            rs485_speed = res;
            return 1;
        default:    
            return 0;
    }
}

static int parse_cmd_line(int argc, char **argv)
{
    int c;
        /* NOTE: chars 'C' and 'M' were excluded! Don't reuse them */
    while(-1 != (c = getopt(argc, argv, "ADEFGHIJKLQRTWYZi:s:b:t:d:l:wvh"))) {
        switch(c) {
            case 'L':
                if(!set_op_mode(opm_list)) return 0;
                break;
            case 'Q':
                if(!set_op_mode(opm_query)) return 0;
                break;
            case 'Z':
                if(!set_op_mode(opm_zero)) return 0;
                break;
            case 'D':
                if(!set_op_mode(opm_dump_eeprom)) return 0;
                break;
            case 'T':
                if(!set_op_mode(opm_setid)) return 0;
                break;
            case 'G':
                if(!set_op_mode(opm_set_conf)) return 0;
                break;
            case 'H':
                if(!set_op_mode(opm_set_hwconf)) return 0;
                break;
            case 'A':
                if(!set_op_mode(opm_action)) return 0;
                action_arg = argv[optind];
                optind++;
                break;
            case 'R':
                if(!set_op_mode(opm_set_reactions)) return 0;
                break;
            case 'W':
                if(!set_op_mode(opm_write_eeprom)) return 0;
                break;
            case 'E':
                if(!set_op_mode(opm_restore_from_eeprom)) return 0;
                break;
            case 'F':
                if(!set_op_mode(opm_restore_factory)) return 0;
                break;
            case 'Y':
                if(!set_op_mode(opm_write_e_reg)) return 0;
                break;
#if 0
            case 'M':
                if(!set_op_mode(opm_message)) return 0;
                break;
            case 'C':
                if(!set_op_mode(opm_intercept_com)) return 0;
                break;
#endif
            case 'K':
                if(!set_op_mode(opm_clear_messagebuf)) return 0;
                break;
            case 'I':
                if(!set_op_mode(opm_input_stream)) return 0;
                break;
            case 'J':
                if(!set_op_mode(opm_reset_1wire)) return 0;
                break;
            case 'i': {
                char *err;
                int i = strtol(optarg, &err, 16);
                if(*optarg != '\0' && *err == '\0') {
                    device_id = i;
                } else {
                    fprintf(stderr, "Invalid device id\n");
                    return 0;
                }
                break;
            }
            case 's':
                use_rs485 = 1;
                rs485_port = optarg;
                warn_nonroot = 0;
                break;
            case 'b':
                if(!extract_baudrate(optarg)) {
                    fprintf(stderr, "Invalid baudrate %s\n", optarg);
                    return 0;
                }
                break;
            case 't': {
                char *err;
                int i = strtol(optarg, &err, 10);
                if(*optarg != '\0' && *err == '\0') {
                    unc0xx_usb_timeout = i;
                } else {
                    fprintf(stderr, "Invalid timeout value\n");
                    return 0;
                }
                break;
            }
            case 'd':
                if(!extract_delay_value(optarg))
                    return 0;
                break;
            case 'l':
                if(optarg[1] || optarg[0] < '0' || optarg[0] > '9') {
                    fprintf(stderr, "Invalid location value\n");
                    return 0;
                }
                action_loc = optarg[0] - '0' + 0x10;
                break;
            case 'w':
                warn_nonroot = 0;
                break;
            case 'v':
                be_verbose++;
                break;
            case 'h':
                help();
                return 0;
            default:
                return 0;
        }
    }
    rest_args = argv + optind;
    if(operation_mode == opm_undef) {
        fprintf(stderr, "You must specify the operation mode\n");
        return 0;
    }
    if(operation_mode != opm_action && action_delay != 0) {
        fprintf(stderr, "The -d option can be used with -A only\n");
        return 0;
    }
    return 1;
}

static void print_devices_header()
{
    printf("<id>  [serial number ]   version\n");

}

static void print_devices_entry(const int sernum[3], int version)
{
    printf("%04x  [%04x:%04x:%04x]  %8d %s\n",
           sernum[2], sernum[0], sernum[1], sernum[2],
           version, version==0  ? "DEPRECATED" : "");
}

static void print_devices(struct unc0xx_scan_item *list)
{
    int n = 0;
    if(!list) {
        printf("No UNC0xx devices found\n");
        return;
    }
    print_devices_header();
    for(; list; list=list->next) {
        print_devices_entry(list->sernum, list->version);
        n++;
    }
    printf("%d UNC0xx device(s) found\n", n);
}

static void verbose_action_print(const char *whatfor, int a)
{
    int i;
    static const char ach[4] = { '=', '+', '-', '^' };
    static const char *act[4] = { "keep  ", "on    ", "off   ", "toggle" };
    printf("   action on %s ", whatfor);
    for(i=3; i>=0; i--)
        printf("%c", ach[a>>(i*2) & 3]);
    printf("    ");
    for(i=3; i>=0; i--)
        printf("%d:%s ", i, act[a>>(i*2) & 3]);
    printf("\n");
}

static const char *line_mode(int line, int hwconf, int *lnn)
{
    int m;
    if(line < 0 || line >= 12) return NULL;
    if(line < 4) { *lnn = line; return "out"; }
    m = (hwconf >> (line / 4 - 1) * 4) & 0x0f;
    switch(m) {
        case 0x0F: { *lnn = line; return "out"; }
        case 0x04: { *lnn = line - 4; return "in"; }
        default:   return NULL;
    }
}

static void display_1w_temp(int temp)
{
    int grad, migrad;
    char tsign = '+';
    switch(temp) {
    case (short)0x8000: /* no temperature here */
        printf("\n");
        break;
    case 0x7fff: /* crc error */
        printf("    temperature: CRC error\n");
        break;
    default:
        grad = temp >> 4;
        migrad = (temp & 0x0f) * 625;
        if(temp < 0) {
            tsign = '-';
            if(migrad > 0) {
                migrad = 10000 - migrad;
                grad = -grad - 1;
            } else {
                grad = -grad;
            }
        }
        printf("    temp: (0x%04hx) %c%d.%04d C\n",
                         temp, tsign, grad, migrad);
    }

}

static void display_1w_2413(int data)
{
    static const char * const labels[4] =
        { " on  ", " err ", "short", " off " };
    printf("   state: (0x%04hx) [%s] [%s] %s\n",
           data, labels[data & 0x03], labels[(data >> 2) & 0x03],
           ((data & 0x0f) == (~(data >> 4) & 0x0f)) ? "OK" : "error");
}

static void display_1w_device(uint8_t *id, int temp)
{
    int j;
    if(!id[0])
        return;
    printf("1w: ");
    for(j=7; j>=0; j--)
        printf("%02x", id[j]);
    switch(id[0]) {
    case 0x10:
    case 0x28:  /* temperature sensors */
        display_1w_temp(temp);
        break;
    case 0x3a:  /* DS2413 */
        display_1w_2413(temp);
        break;
    default:
        printf("\n");
    }
}

static void display_xstate_legacy(struct unc0xx_extra_state_str_common *ex)
{
    int i;
    struct unc0xx_extra_state_str_legacy *extra =
        (struct unc0xx_extra_state_str_legacy*) ex;
    if(extra->data_buffer[0] & 0x07) {
        for(i = 1; i < (extra->data_buffer[0] & 0x07); i++)
            printf("%02x ", extra->data_buffer[i]);
        printf("\n");
    }
    for(i=0; i<4; i++) {
        display_1w_device(extra->ow_ids[i], extra->ow_temperature[i]);
    }
}

static void display_xstate_onewire(struct unc0xx_extra_state_str_common *ex)
{
    int i;
    struct unc0xx_extra_state_str_onewire *extra =
        (struct unc0xx_extra_state_str_onewire*) ex;
    for(i=0; i<6; i++) {
        display_1w_device(extra->ow_ids[i], extra->ow_temperature[i]);
    }
}

static void display_xstate_delays(struct unc0xx_extra_state_str_common *ex)
{
    int i, tm;
    struct unc0xx_extra_state_str_delays *extra =
        (struct unc0xx_extra_state_str_delays*) ex;
    for(i=0; i<UNC_MAX_DELAYED_ACTIONS; i++) {
        if(extra->actions[i].grp > 0x0e)
            continue;
        tm = extra->counters[i];
        printf("delayed action %d: %2x%02x in %8d cycles (", i,
               extra->actions[i].grp, extra->actions[i].act, tm);
        if(tm >= 360000) {
            printf("%dh ", tm / 360000);
            tm %= 360000;
        }
        if(tm >= 6000) {
            printf("%dm ", tm / 6000);
            tm %= 6000;
        }
        printf("%d.%02ds)\n", tm / 100, tm % 100);
    }
}

static int query_device(struct unc0xx_devhdl *dev_hdl)
{
    struct unc0xx_global_configuration_str conf;
    struct unc0xx_global_state_str stat;
    struct unc0xx_extra_state_str_common extra;
    int first_input, last_input;
    int i, r;

    r = unc0xx_get_configuration(dev_hdl, &conf);
    if(r == -1) {
        fprintf(stderr, "get_configuration failed\n");
        //return 1;
    }
    r = unc0xx_get_fullstate(dev_hdl, &stat);
    if(r == -1) {
        fprintf(stderr, "get_fullstate failed\n");
        return 1;
    }

    first_input = ((conf.hardware_config & 0x0f) == 0x04) ? 0 : 4;
    last_input  = ((conf.hardware_config & 0xf0) == 0x40) ? 7 : 3;

    printf("version: %d\n"
           "hardware configuration: %08x user configuration: %08x\n",
           conf.version, conf.hardware_config, conf.user_config);
    printf("state register: %04x   fsm state: %04x",
           stat.state_bits, stat.fsm_state);
    if(conf.version >= 2120203)
        printf("  timemark: %09d\n", stat.timemark);
    else
        printf("\n");
    if(be_verbose) {
        printf("   ");
        for(i=11; i>=0; i--) {
            int lnn;
            const char *lm = line_mode(i, conf.hardware_config, &lnn);
            if(!lm) continue;
            int s = ((stat.state_bits>>(i > 3 ? i + 4 : i))&1);
            printf("%s%d:%s ", lm, lnn, s ? "on ":"off");
            if(i && !(i%4)) printf("\n   ");
        }
        printf("\n");
    }
    for(i = first_input; i<=last_input; i++)
    {
        if(conf.reactions[i] & 0x80000000) /* analog mode */
            continue;
        printf("in%d: mode %08x  counter %d\n",
                i, conf.reactions[i], stat.counters[i]);
        if(be_verbose) {
            static const char *icm[] =
                { "none", "press", "release", "press&release" };
            printf("   increment counter on: %s\n",
                   icm[(conf.reactions[i]>>29)&3]);
            verbose_action_print("release:", (conf.reactions[i]>>8)&0xff);
            verbose_action_print("press:  ", (conf.reactions[i])   &0xff);
        }
    }
    if(stat.buf_start != stat.buf_end) {
        printf("message: <[");
        stat.buf_end &= 0x1f;
        stat.buf_start &= 0x1f;
        while(stat.buf_start != stat.buf_end) {
            putchar(stat.message_buffer[stat.buf_start]);
            ++stat.buf_start;
            stat.buf_start &= 0x1f;
        }
        printf("]>\n");
    } 
    if(conf.version >= 2120203) {
        int ni = 0;
        do {
            r = unc0xx_get_extrastate(dev_hdl, &extra);
            if(r == -1) {
                fprintf(stderr, "get_extrastate failed\n");
                return 1;
            }
            if(be_verbose) {
                printf("page_seq=%02x page_version=%02x reserved1=%02x\n",
                       extra.page_seq, extra.page_version, extra.reserved1);
            }
            switch(extra.page_version) {
            case uncxstate_page_legacy: /* legacy */
                display_xstate_legacy(&extra);
                break;
            case uncxstate_page_onewire: /* contemporary versions */
                display_xstate_onewire(&extra);
                break;
            case uncxstate_page_delays: /* delays: since 4xxxxxx */
                display_xstate_delays(&extra);
                break;
            default:
                fprintf(stderr, "extra_state version %d not supported!\n",
                        extra.page_version);
            }
        } while((extra.page_seq & 0x80) && ++ni < 16);
    }
    return 0;
}

static int zero_query_counters(struct unc0xx_devhdl *dev_hdl)
{
    int cnt[UNC_INPUTS_COUNT];
    int r, i;
    r = unc0xx_zero_query_counters(dev_hdl, cnt, UNC_INPUTS_COUNT);
    if(r < 0) {
        fprintf(stderr, "error: %s\n", unc0xx_get_errormsg(dev_hdl));
        return 1;
    }
    for(i=0; i<r; i++) {
        printf("%d ", cnt[i]);
    }
    printf("\n");
    return 0;
}

static void hexdump(char *buffer, int len, int base)
{
    int i;
    for(i = 0; i < len; i++){
        if(i % 16 == 0){
            if(i != 0) printf("\n");
            printf("%03x:  ", base + i);
        } else
        if(i % 8 == 0) {
            printf("   ");
        } else
        if(i % 4 == 0) {
            printf("  ");
        } else {
            printf(" ");
        }
        printf("%02x", buffer[i] & 0xff);
    }
    if(i != 0)
        printf("\n");
}

static int dump_eeprom(struct unc0xx_devhdl *dev_hdl, int version)
{
    char buf[128];
    int r, tot;
    for(tot = 0; tot < ((version >= 2120203) ? 512 : 128); ) {
        r = unc0xx_dump_eeprom(dev_hdl, buf);
        if(r <= 0)
            return 1;
        hexdump(buf, r, tot);
        tot += r;
    }
    return 0;
}

static int generic_set_command(struct unc0xx_devhdl *dev_hdl,
                               char opt, int check,
                               int (*fun)(struct unc0xx_devhdl*, unsigned int))
{
    char *err;
    int r;
    long id;
    if(check) {
        int ok = 1;
        if(!rest_args || !rest_args[0] || !rest_args[1])
            ok = 0;
        else
            ok = (strcmp(rest_args[1], "really") == 0);
        if(!ok) {
            fprintf(stderr, "Don't use -%c. "
                "The device can be physically damaged with this option.\n"
                "If you know what you do, specify the word 'really'\n",
                opt);
            return 1;
        }
    }
    if(!rest_args || !*rest_args) {
        fprintf(stderr, "No number specified for -%c\n", opt);
        return 1;
    }
    id = strtol(*rest_args, &err, 16);
    if(*rest_args != '\0' && *err == '\0') {
        /* okay */
        r = fun(dev_hdl, id);
        return r==0 ? 0 : 1;
    } else {
        fprintf(stderr, "Invalid parameter (must be a hex number)\n");
        return 1;
    }
}

static int set_new_id(struct unc0xx_devhdl *dev_hdl)
{
    return generic_set_command(dev_hdl, 'T', 0, unc0xx_set_new_id);
}

static int set_config(struct unc0xx_devhdl *dev_hdl)
{
    return generic_set_command(dev_hdl, 'G', 0, unc0xx_set_configuration);
}

static int set_hwconfig(struct unc0xx_devhdl *dev_hdl)
{
    return generic_set_command(dev_hdl, 'H', 1, unc0xx_set_hardware_conf);
}

static int extract_action_bin(char *databuf, uint16_t *res)
{
    int action = 0;
    while(*databuf) {
        action *= 2;
        switch(*databuf) {
        case '1':
            action += 1;
            /* intentionally no break here */
        case '0':
            break;
        default:
            return 0;
        }
        databuf++;
    }
    *res = action & 0xff;
    return 1;
}

static int extract_action_hex(char *databuf, uint16_t *res)
{
#if 0
    int d1, d2;
    d1 = hexdig2num(databuf[0]);
    if(d1 != -1) {
        d2 = hexdig2num(databuf[1]);
        if(d2 != -1 && databuf[2] == '\0') {
            *res = d1<<4 | d2;
            return 1;
        }
    }
    return 0;
#else
    char *err;
    *res = strtol(databuf, &err, 16);
    return *res && !*err;
#endif
}

static int extract_action_9(char *databuf, uint16_t *res)
{
    int ereg;
    char *err;
    if(databuf[2] != '@')
        return 0;
    if(databuf[0] != '+' && databuf[0] != '-' && databuf[0] != '.')
        return 0;
    if(databuf[1] != '+' && databuf[1] != '-' && databuf[1] != '.')
        return 0;
    if(databuf[3] == 'x' || databuf[3] == 'X')
        ereg = strtol(databuf+4, &err, 16);
    else
        ereg = strtol(databuf+3, &err, 10);
    if(*err)
        return 0;
    if(ereg < 0 || ereg > 47)
        return 0;
    *res = 0x0900 | (ereg << 2) |
        (databuf[0] == '+' ? 2 : 0) |
        (databuf[1] == '+' ? 1 : 0) ;
    return 1;
}

static int extract_action_mnem(char *databuf, uint16_t *res)
{
    int action = 0;
    char *datap = databuf;
    while(*datap) {
        action <<= 2;
        switch(*datap) {
        case '=':
            break;       /* 00 == keep untouched */
        case '^':
            action |= 3; /* 11 == switch state */
            break;
        case '+':
            action |= 1; /* 01 == set on */
            break;
        case '.':
        case '-':
            action |= 2; /* 10 == set off */
            break;
        case '@':
            return extract_action_9(databuf, res);
        default:
            return 0;
        }
        datap++;
    }
    *res = action & 0xff;
    return 1;
}

static int extract_action(char *databuf, uint16_t *res)
{
    switch(databuf[0]) {
        case '0':
        case '1':
            return extract_action_bin(databuf, res);
        case 'x':
        case 'X':
            return extract_action_hex(databuf+1, res);
        case '.':
        case '+':
        case '-':
        case '^':
        case '=':
            return extract_action_mnem(databuf, res);
        default:
            return 0;
    }
}


static int perform_action(struct unc0xx_devhdl *dev_hdl, int version)
{
    uint16_t action = 0; /* 0 is to make stupid compiler happy */
    uint16_t agrp;
    int r;
    char *err;
    if(!action_arg) {
        fprintf(stderr, "No action specified for -A\n");
        return 1;
    }
    if(strchr(action_arg, ':')) {
        char *p = strrchr(action_arg, ':');
        if(!extract_action(p+1, &action)) {
            fprintf(stderr, "Invalid action\n");
            return 1;
        }
        *p = 0;
        agrp = strtoul(action_arg, &err, 16) << 8;
        if(*action_arg == '\0' || *err != '\0') {
            fprintf(stderr, "Invalid action group\n");
            return 1;
        }
        action &= 0x0ff;
        action |= agrp;
    } else {
        if(!extract_action(action_arg, &action)) {
            fprintf(stderr, "Invalid action\n");
            return 1;
        }
    }
    if(version < 2130914) {
        fprintf(stderr, "NOTICE The firmware version is so old that is uses "
                        "obsolete action codes\n");
        if(action > 0x200 && action < 0x1000) {
            fprintf(stderr, "WARNING I'll try with your action code but "
                            "the device will not understand\n");
        }
    } else {
        if(action >= 0x1000) {
            fprintf(stderr, "WARNING You seem to be using obsolete action "
                            "code while the firmware is newer than 2130914\n");
        }
    }
    if(be_verbose)
        fprintf(stderr, "NOTE: action %04x\n", action);
    if(action_delay == 0) {
        r = unc0xx_perform_action(dev_hdl, action);
    } else {
        if(version < 4000000) {
            fprintf(stderr, "WARNING your device is too old for the "
                            "delayed action feature\n");
        }
        r = unc0xx_delayed_action(dev_hdl, action, action_delay, action_loc);
    }
    if(r != 0) {
        fprintf(stderr, "error: %s\n", unc0xx_get_errormsg(dev_hdl));
        return 1;
    }
    return 0;
}

static int set_reactions(struct unc0xx_devhdl *dev_hdl)
{
    char *err;
    int r;
    int line;
    unsigned long reaction;

    if(!rest_args || !*rest_args) {
        fprintf(stderr, "No input line number specified for -R\n");
        return 1;
    }
    if(rest_args[0][1]!='\0' || rest_args[0][0]>'7' || rest_args[0][0]<'0'){
        fprintf(stderr, "Invalid input line (must be 0, 1, ..., 7)\n");
        return 1;
    }
    line = rest_args[0][0] - '0';
    if(!rest_args[1]) {
        fprintf(stderr, "No reaction value specified for -R\n");
        return 1;
    }
    if(strchr(rest_args[1], ':')) {
        char *p = strrchr(rest_args[1], ':');
        uint16_t ac_press, ac_release;
        if(!extract_action(p+1, &ac_press)) {
            fprintf(stderr, "Invalid action string %s\n", p+1);
            return 1;
        }
        *p = 0;
        p = strrchr(rest_args[1], ':');
        if(!extract_action(p ? p+1 : rest_args[1], &ac_release)) {
            fprintf(stderr, "Invalid action string %s\n",
                            p ? p+1 : rest_args[1]);
            return 1;
        }
        if(p) {
            *p = 0;
            reaction = strtoul(rest_args[1], &err, 16) << 16;
            if(*rest_args[1] == '\0' || *err != '\0') {
                fprintf(stderr, "Invalid reaction value\n");
                return 1;
            }
        } else {
            // only one ':' means there's no flags part, should use 0x6000 
            reaction = 0x60000000;
        }
        reaction |= ((unsigned long) ac_press & 0xff);
        reaction |= (((unsigned long) ac_release) & 0xff) << 8;
    } else {
        reaction = strtoul(rest_args[1], &err, 16);
        if(*rest_args[1] == '\0' || *err != '\0') {
            fprintf(stderr, "Invalid reaction value\n");
            return 1;
        }
    }
    if(!(reaction & 0x60000000)) {
        fprintf(stderr, "Warning: counter disabled\n");
    }
    r = unc0xx_set_reactions(dev_hdl, line, reaction);
    if(r != 0) {
        fprintf(stderr, "error: %s\n", unc0xx_get_errormsg(dev_hdl));
        return 1;
    }
    return 0;
}

static int write_e_reg(struct unc0xx_devhdl *dev_hdl)
{
    char *err, *p;
    int r, i;
    int regnum;
    char data[8];

    if(!rest_args || !*rest_args) {
        fprintf(stderr, "No E reg number specified for -Y\n");
        return 1;
    }
    regnum = **rest_args == 'x' ? 
        strtol(*rest_args + 1, &err, 16) :
        strtol(*rest_args, &err, 10);
    if(!**rest_args || *err || regnum < 0 || regnum > 47){
        fprintf(stderr, "Invalid E register number, must be 0..47\n");
        return 1;
    }
    if(!rest_args[1]) {
        fprintf(stderr, "No data specified for -Y\n");
        return 1;
    }
    p = rest_args[1];
    for(i=0; i<8; i++) {
        while(*p == ' ') p++;
        int d1 = hexdig2num(*p++);
        while(*p == ' ') p++;
        int d2 = hexdig2num(*p++);
        if(d1 == -1 || d2 == -1) {
            fprintf(stderr, "Invalid data for -Y\n");
            return 1;
        }
        data[7-i] = (d1 << 4) | d2;
    }
    r = unc0xx_write_e_reg(dev_hdl, regnum, data);
    if(r != 0) {
        fprintf(stderr, "error: %s\n", unc0xx_get_errormsg(dev_hdl));
        return 1;
    }
    return 0;
}

static int input_stream(struct unc0xx_devhdl *dev_hdl)
{
    int rc;
    char buf[8];
    if(unc0xx_takeover_stream(dev_hdl) < 0) {
        fprintf(stderr, "Error taking over the stream: %s\n",
                        unc0xx_get_errormsg(dev_hdl));
        return 1;
    }
    while((rc = unc0xx_read_stream(dev_hdl, buf)) > 0) {
        int i;
        for(i=0; i<rc; i++)
            putchar(buf[i]);
        fflush(stdout);
    }
    return rc == 0 ? 0 : 1;
}

extern int intercept_com(struct unc0xx_devhdl *dev_hdl);

int main(int argc, char **argv)
{
    struct unc0xx_devhdl* dev_hdl = NULL;
    int version;

    if(argc<2) {
        help();
        return 1;
    }

    if(!parse_cmd_line(argc, argv)) {
        fprintf(stderr, "Use -h for help\n");
        return 1;
    }

    if(use_rs485) {
        if(device_id == 0)
            device_id = 0x7775;
#if 1
        dev_hdl = unc0xx_rs485_probe(rs485_port, rs485_speed,
                                     device_id, &version);
        if(!dev_hdl) {
            perror(rs485_port);
            return 1;
        }
        if(version <= 0) {
            fprintf(stderr, "device %04x: %s\n",
                            device_id, unc0xx_get_errormsg(dev_hdl));
            return 1;
        }
        if(operation_mode == opm_list) {
            int sernum[3] = { 0, 0, 0 };
            sernum[2] = device_id;
            print_devices_header();
            print_devices_entry(sernum, version);
            return 0;
        }
#else
        dev_hdl = unc0xx_rs485_open(rs485_port, device_id);
        if(!dev_hdl) {
            perror(rs485_port);
            return 1;
        }
#endif
    } else {
        struct unc0xx_scan_item *list;
        unc0xx_usb_init();
        list = unc0xx_scan(device_id);

#ifndef WIN32
        if(warn_nonroot && geteuid() != 0 && !list) {
            fprintf(stderr,
                   "WARNING: usually only root can operate this program\n");
        } 
#endif

        if(operation_mode == opm_list) {  /* very special case */
            print_devices(list);
            return list ? 0 : 1;
        }
        if(!list) {
            fprintf(stderr, "No devices found\n");
            return 1;
        }
        if(device_id == 0 && list->next) {
            fprintf(stderr,
                "WARNING: several devices found, using device %04x.\n",
                list->sernum[2]);
            fprintf(stderr,
                "It is strongly recommended to use -i to select the device\n");
        }
        dev_hdl = list->dev_hdl;
        list->dev_hdl = NULL;
        version = list->version;
        unc0xx_destroy_list(list);
    }


    if(!dev_hdl) {
        fprintf(stderr, "The device is incompatible, check versions\n");
        return 1;
    }
    switch(operation_mode) {
        case opm_undef:
        case opm_list:
            return 1; /* never happens */
        case opm_query:
            return query_device(dev_hdl);
        case opm_zero:
            return zero_query_counters(dev_hdl);
        case opm_dump_eeprom:
            return dump_eeprom(dev_hdl, version);
        case opm_setid:
            return set_new_id(dev_hdl);
        case opm_set_conf:
            return set_config(dev_hdl);
        case opm_set_hwconf:
            return set_hwconfig(dev_hdl);
        case opm_action:
            return perform_action(dev_hdl, version);
        case opm_set_reactions:
            return set_reactions(dev_hdl);
        case opm_write_eeprom:
            return unc0xx_save_to_eeprom(dev_hdl)==0?0:1;
        case opm_restore_from_eeprom:
            return unc0xx_restore_from_eeprom(dev_hdl)==0?0:1;
        case opm_restore_factory:
            return unc0xx_restore_factory(dev_hdl)==0?0:1;
        case opm_write_e_reg:
            return write_e_reg(dev_hdl);
        case opm_clear_messagebuf:
            return unc0xx_clear_msg(dev_hdl)==0?0:1;
        case opm_reset_1wire:
            return unc0xx_reset_1wire(dev_hdl)==0?0:1;
        case opm_input_stream:
            if(version < 2120131) {
                fprintf(stderr, "Device's firmware is too old. ");
                fprintf(stderr, "It doesn't support this mode.\n");
                fprintf(stderr, "Consider updating it\n");
                return 1;
            } else {
                return input_stream(dev_hdl);
            }
    }
    return 0;
}
