/*
 * UNC0xx firmware. (c) Unicontrollers, Ltd., 2013-2014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file rs485.c
 * Rs485 handling
 *
 * Written by Andrey V. Stolyarov
 *
 */


#define UART_SOME_ERROR \
    (UART_NO_DATA|UART_FRAME_ERROR|UART_OVERRUN_ERROR|UART_BUFFER_OVERFLOW)

#include <avr/eeprom.h>

#include "uart.h"

#include "ucd.h"
#include "unc_prot.h"
#include "control.h"
#include "unc_prim.h"
#include "misc.h"

#include "rs485.h"

/*
    Please note what we implement here is a PSEUDO ascii modbus protocol.
    This SHOULD be no problem for other devices provided that the addresses
    of the devices that YOU setup do not conflict.  Please note we use
    16-bit addressing (the User part of serial number is used as the
    device's address), MSB first, with an additional digit for each byte.
    So you can, e.g., setup them as 9001, 9002, 9003, 9004... -- and relax,
    because regular modbus never uses 0x90 as a device's address.
 */

#define DIR_OUTPUT() (PORTC |= (1<<5))
#define DIR_INPUT() (PORTC &= ~(1<<5))


/* ------------ some older initialization variants, all disabled ----- */
#if 0

#ifndef F_CPU
#define F_CPU 12000000UL
#endif

#define UART_BAUD_RATE      19200
//#define UART_BAUD_RATE      9600
    uart_init(UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU)); 

    static uint8_t baudsels[8] = {
        UART_BAUD_SELECT(19200, F_CPU),
        UART_BAUD_SELECT(4800,  F_CPU),
        UART_BAUD_SELECT(9600, F_CPU),
        UART_BAUD_SELECT(19200, F_CPU),
        UART_BAUD_SELECT(38400, F_CPU),
        UART_BAUD_SELECT(57600, F_CPU),
        UART_BAUD_SELECT(115200, F_CPU)
    };


    static uint8_t baudsels[8] = { 38, 155, 77, 25, 19, 12, 6, 3 };
    uart_init(baudsels[(global_configuration.user_config_bytes[0]>>4) & 7]);
#endif
/* ------------------------------------------------------------------- */


void init_rs485(void)
{
    uint8_t b = global_configuration.hardware_config_bytes[1];
    if(!b)
        b = 38; /* 19200 */
    uart_init(b);
}

enum pmbstates {
    pmb_wait,        /* waiting for ':' */
    pmb_address,     /* receiving station id */
    pmb_body1,       /* receiving the body, expecting first half-byte */
    pmb_body2,       /* receiving the body, expecting second half-byte */
    pmb_body_check,  /* receiving the body, expecting parity half-byte */
    pmb_wait_ret_ans,/* waiting for newline to send the answer */
    pmb_answer,      /* waiting the time, then sending the answer */
    pmb_pause,       /* waiting for tx buffer to empty off */
    pmb_pause2,      /* one cycle before switching DIRECTION back to INPUT */
    pmb_ignoring     /* ignoring everything until '\n' */
};

static uint8_t the_state = pmb_wait;
static uint16_t address_from_message;
static uint16_t address_from_eeprom;
static uint8_t current_index;
static uint8_t current_byte;
static int8_t pending_response_len;

static void body_command_start(uint8_t b)
{
    switch(b) {
    case uncreq_command:
    case uncreq_long_command:
        start_incoming_request(b);
        process_incoming_data(&b, 1);
        the_state = pmb_body1;
        break;
    default:
        pending_response_len = start_data_request(b);
        the_state = pmb_wait_ret_ans;
    }
}

static void send_hex_byte(uint8_t b)
{
    uart_putc(hexdigit(b >> 4));
    uart_putc(hexdigit(b & 0x0f));
}

static void send_hex_byte_check(uint8_t b)
{
    send_hex_byte(b);
    uart_putc(hexdigit((b ^ (b >> 4)) & 0x0f));
}

static void send_the_answer()
{
    uint8_t b;

    uart_putc(':');
    send_hex_byte(address_from_message >> 8);
    send_hex_byte(address_from_message & 0xff);
    for(b = 0; b < pending_response_len; b++)        
        send_hex_byte_check(transfer_data_pointer[b]);
    uart_putc('\n');
    the_state = pmb_pause;
}

static void rs485_process_char(uint8_t c)
{
    int8_t digv = hexdig2num(c);
    if(c == '\r')  /* simply ignored, in all states */
        return;
    if(c == '\n') {
        if(the_state == pmb_wait_ret_ans) {
            the_state = pmb_answer;
            DIR_OUTPUT();
            send_the_answer();
        } else
            the_state = pmb_wait;
        return;
    }
    if(c == ':') {
        the_state = pmb_address;
        address_from_message = 0;
        current_index = 0;
        return;
    }
    if(the_state == pmb_wait) {
        return;
    }
    if(digv == -1) {
        the_state = pmb_ignoring;
        return;
    }
    switch(the_state) {
    case pmb_address: 
        address_from_message <<= 4;
        address_from_message |= digv;
        if(++current_index == 4) {
            eeprom_read_block(&address_from_eeprom, (uint8_t*)4, 2);
            if(address_from_message == address_from_eeprom) {
                the_state = pmb_body1;
                current_index = 0;
                // global_state_putchar('*');
            } else {
                the_state = pmb_ignoring;
                // global_state_putchar('@');
            }
        }
        break;
    case pmb_body1: 
        current_byte = digv << 4;
        the_state = pmb_body2;
        break;
    case pmb_body2: 
        current_byte |= digv;
        the_state = pmb_body_check;
        break;
    case pmb_body_check:
        if((current_byte >> 4) ^ (current_byte & 0x0f) ^ digv) {
           /* their xor must be zero, but it is not */
           the_state = pmb_ignoring;
           break;
        }
        if(current_index++ == 0) {
            body_command_start(current_byte); /* will change the state */
            // global_state_putchar('^');
        } else {
            if(process_incoming_data(&current_byte, 1)) {
                // global_state_putchar('#');
                the_state = pmb_ignoring;
                pending_response_len = 1;
                /* this is damn hack as we use the variable which doesn't
                   belong us (transfer_data_pointer) */
                transfer_data_pointer = &current_byte;
                current_byte = 0;
                the_state = pmb_wait_ret_ans;
            } else {
                the_state = pmb_body1;
            }
        }
        break;
    default:
        ;
    }
}

void process_serial_input()
{
    uint16_t c;
    while(((c=uart_getc()) & UART_SOME_ERROR) == 0) {
        rs485_process_char(c);
    }
}

void rs485_hook()
{
    if(the_state == pmb_pause2) {
        DIR_INPUT();
        the_state = pmb_wait;
    }
    if(the_state == pmb_pause && uart_txbuf_empty())
        the_state = pmb_pause2;
}
