/*
 * UNC0xx GUI-based software. (c) Unicontrollers, Ltd., 20122014
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file config.cpp
 * Manages the host-side configuration for devices, such as names,
 * line descriptions, visibility etc.
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifndef _WIN32
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#endif

#include "inifile.hpp"

#include "config.hpp"


static IniFileParser *config = 0;
static char *config_fname = 0;

struct conf_item {
    DeviceConfig *dc;
    conf_item *next;
};

static conf_item* first_dc = 0;


bool InitConfig(const char *fname)
{
    if(config) return false;
    config = new IniFileParser;
    if(fname) {
        config_fname = strdup(fname);
    } else {
        // okay, the harder case: we need to find it

        // first, let's try uncmonitor.rc in the current directory
        config_fname = strdup("uncmonitor.rc");
        // all the following are Unix only, so in case of Windows
        // we just leave the things where they are

#ifndef _WIN32
        struct stat s;
        const char *home;
        if(-1 != stat(config_fname, &s))
            goto found;

        free(config_fname);
        // next possibility is ~/.uncmonitor.rc
        home = getenv("HOME");
        if(home) {
            static const char umon[] = "/.uncmonitor.rc";
            config_fname = (char*)malloc(strlen(home) + sizeof(umon));
            strcpy(config_fname, home);
            strcat(config_fname, umon);
            if(-1 != stat(config_fname, &s))
                goto found;
        }

        // the last possibility is "/etc/uncmonitor.rc" but if it
        // isn't found, then we fall back to home dir
        
        static const char etcmon[] = "/etc/uncmonitor.rc";
        if(-1 != stat(etcmon, &s)) {
            free(config_fname);
            config_fname = strdup(etcmon);
        }
    found:
        ;
#endif
    }
    return config->Load(config_fname);
}

DeviceConfig* GetDeviceConfig(int id)
{
    conf_item *t;
    for(t = first_dc; t; t = t->next)
        if(t->dc->GetId() == id) return t->dc;
    // not found, create new
    t = new conf_item;
    t->next = first_dc;
    t->dc = new DeviceConfig(id);
    first_dc = t;
    return t->dc;
}

static bool SaveConfig()
{
    if(!config) return false;
    return config->Save(config_fname);
}

static char defaultbuf[64];
static char lnumbuf[32];

DeviceConfig::DeviceConfig(int aid)
    : id(aid)
{
    sprintf(strid, "%04x", aid);
}

const char *DeviceConfig::GetName() const
{
    sprintf(defaultbuf, "Device #%04x", id);
    if(!config) return defaultbuf;
    return config->GetTextParameter("Device", strid, "Name", defaultbuf);
}

device_visibility DeviceConfig::GetVisibility() const
{
    if(!config) return devv_default;
    const char *s = config->GetTextParameter("Device", strid, "Show", "");
    if(strcasecmp(s, "Ignore")==0) return devv_ignored;
    if(strcasecmp(s, "Never") ==0) return devv_hidden;
    if(strcasecmp(s, "Always")==0) return devv_always_visible;
    return devv_default;
}

const char *DeviceConfig::GetLineName(int n) const
{
    line_type lt = GetLineType(n);
    sprintf(defaultbuf, "%sput #%d",
         lt == lnt_output ? "Out" : "In",
         lt == lnt_input  ? n - 4 : n);
    if(!config) return defaultbuf;
    sprintf(lnumbuf, "%d", n);
    return config->GetModifiedTextParameter("Device", strid, "Line",
                                            lnumbuf, defaultbuf);
}


line_type DeviceConfig::GetModuleType(int n) const
{
    if(n < 0 || n > 1) return lnt_none;
    sprintf(lnumbuf, "Module%d", n);
    const char *val =
        config->GetModifiedTextParameter("Device", strid, lnumbuf,
                                         "seen", "none");
    if(!val || strcasecmp(val, "none") == 0)
        return lnt_none;
    if(strcasecmp(val, "input") == 0)
        return lnt_input;
    if(strcasecmp(val, "output") == 0)
        return lnt_output;
    return lnt_none;
}

line_type DeviceConfig::GetLineType(int n) const
{
    if(n < 4) return lnt_output;
    return GetModuleType((n - 4) / 4);
}

void DeviceConfig::SetName(const char *name)
{
    if(!config) return;
    config->SetTextParameter("Device", strid, "Name", name);
}

void DeviceConfig::SetLineName(int n, const char *name)
{
    if(!config) return;
    sprintf(lnumbuf, "Line:%d", n);
    config->SetTextParameter("Device", strid, lnumbuf, name);
}

static const char *line_type_name(line_type t)
{
    switch(t) {
        case lnt_none: return "none";
        case lnt_input: return "input";
        case lnt_output: return "output";
        default: return "none";
    }
}

void DeviceConfig::SetModuleType(int n, line_type t, bool seen)
{
    
    sprintf(lnumbuf, "Module%d%s", n, seen ? ":seen" : "");
    config->SetTextParameter("Device", strid, lnumbuf, line_type_name(t));
}

static const char *dev_vis_name(device_visibility dv)
{
    switch(dv) {
    case devv_always_visible: return "Always";
    case devv_default:        return "Default";
    case devv_hidden:         return "Never";
    case devv_ignored:        return "Ignore";
    }
}

void DeviceConfig::SetVisibility(device_visibility dv)
{
    config->SetTextParameter("Device", strid, "Show", dev_vis_name(dv));
}

bool DeviceConfig::Save()
{
    return SaveConfig();
}

ForciblyShownDevices::ForciblyShownDevices()
{
    if(!config) {
        count = 0;
        array = 0;
    } else {
        int c = config->GetSectionCount("Device");
        array = new int[c];
        count = 0;
        for(int i = 0; i<c; i++) {
            const char *sn = config->GetSectionName("Device", i);
            const char *vis = config->GetParam("Device", sn, "Show");
            if(vis && strcasecmp(vis, "Always") == 0) {
                char *endptr;
                long id = strtol(sn, &endptr, 16);
                if(*sn && !*endptr)
                    array[count++] = id;
            }
        }
    }
}

ForciblyShownDevices::~ForciblyShownDevices()
{
    if(array)
        delete [] array;
}
