/*
 * UNC0xx GUI-based software. (c) Unicontrollers, Ltd., 2012
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file hwcfgdlg.cpp
 * Implementation of the "Device (hardware) configuration" dialog
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <wx/dialog.h>
#include <wx/msgdlg.h>
#include <wx/textdlg.h>
#include <wx/gbsizer.h>
#include <wx/checkbox.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/button.h>
#include <wx/choice.h>

#include "config.hpp"
#include "glue.hpp"
#include "gui.hpp"

#include "hwcfgdlg.hpp"

struct reg_bit {
    unsigned long mask;
    const char *descr;
};

static const reg_bit userconf_reg[] = {
    { 0x800000, "Scan 1-Wire bus" },
    { 0x400000, "Output 1-Wire IDs on connection" },
    { 0x200000, "Output 1-Wire IDs on disconnection" },
    { 0x100000, "Handle 1-Wire temperature gauges" },
    { 0x080000, "Scan E-regs for iButton keys" },
    { 0x040000, "Scan E-regs for temperature events" },
    { 0x000100, "Enable USB output stream" },
    { 0, 0 }
};

static const reg_bit reaction_reg[] = {
    { 0x40000000, "Increment counter when looped" },
    { 0x20000000, "Increment counter when unlooped" },
    { 0x10000000, "Keep counter's parity" },
    { 0x08000000, "Output message when looped" },
    { 0x04000000, "Output message when unlooped" },
    { 0, 0 }
};


///////////////////////////////////////////////////////////////////////////
//

class RRegDialog: public wxDialog {
    int id;
    int regn;
    unsigned long value;
    UncDeviceCollection *the_collection;
    class HwConfDialog *the_master;
    wxCheckBox **bits;
    int bits_count;
    wxTextCtrl *actions[2];
    int action_vals[2];
public:
    RRegDialog(int aid, int aregn, unsigned long avalue,
               UncDeviceCollection *a_collection,
               HwConfDialog *a_master);
    ~RRegDialog();

private:
    void OnOk(wxCommandEvent& event);
    void OnCancel(wxCommandEvent& event);
    void OnText(wxCommandEvent& event);

    static wxString MakeCaption(int id, int n);
    void SyncAction(int i);
    void SyncActions();
    void SyncActionHighByte(int from, int to);

    DECLARE_EVENT_TABLE()
};

class HwConfDialog: public wxDialog {
    int id;
    UncDeviceCollection *the_collection;
    DeviceConfig *conf;
    wxBoxSizer *sizer;
    wxCheckBox **bits;
    int bits_count;
    wxTextCtrl *input_delay;
    wxStaticText *react_regs[8];
    wxButton *react_buttons[8];
    wxTextCtrl *dev_id;
    wxChoice *modules[2];
    wxTextCtrl *tact_length;

    RRegDialog *rreg_dialogs[8];

    unsigned long hw, user, reacts[8];
public:
    HwConfDialog(int id, UncDeviceCollection *a_coll);
    ~HwConfDialog();

    void RRegDlgDisappears(int n);
    void Sync();
private:
    static wxString MakeCaption(int id);
    //void SyncLineLabels();
    void OnSubmitUserreg(wxCommandEvent& event);
    void OnSetId(wxCommandEvent& event);
    void OnSetHardware(wxCommandEvent& event);
    void OnSaveToEeprom(wxCommandEvent& event);
    void OnRestoreEeprom(wxCommandEvent& event);
    void OnRestoreFactory(wxCommandEvent& event);
    void OnButton(wxCommandEvent& event);
    void OnQuit(wxCommandEvent& event);

    DECLARE_EVENT_TABLE();
};

///////////////////////////////////////////////////////////////////////////
//

static int make_bits_checkboxes(wxWindow *win, wxBoxSizer *sz,
                                const reg_bit* bits, wxCheckBox ***boxes)
{
    int res = 0;
    while(bits[res].descr) res++;
    *boxes = new wxCheckBox*[res];
    for(int i = 0; i < res; i++) {
        (*boxes)[i] = new wxCheckBox(win, wxID_NONE,
                                     wxString(bits[i].descr, wxConvLocal));
        sz->Add((*boxes)[i]);
    }
    return res;
}

///////////////////////////////////////////////////////////////////////////
//

RRegDialog::RRegDialog(int aid, int aregn, unsigned long avalue,
           UncDeviceCollection *a_collection,
           HwConfDialog *a_master)
    : wxDialog(NULL, -1, MakeCaption(aid, aregn), wxDefaultPosition,
               wxDefaultSize,
               wxCAPTION|wxTHICK_FRAME|wxMINIMIZE_BOX|wxSYSTEM_MENU|
               wxCLOSE_BOX|wxDIALOG_NO_PARENT),
      id(aid), regn(aregn), value(avalue),
      the_collection(a_collection), the_master(a_master)
{
    int i;
    wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);

    bits_count = make_bits_checkboxes(this, sizer, reaction_reg, &bits);
    for(i = 0; i < bits_count; i++) {
        bits[i]->SetValue((value & reaction_reg[i].mask) != 0);
    }

    wxString ac_lab[2] = { wxT("<<ON>>"), wxT("<<OFF>>") };
    for(i=0; i<2; i++) {
        wxBoxSizer *s = new wxBoxSizer(wxHORIZONTAL);
        wxString str = ac_lab[i] + wxT(" event action");
        s->Add(new wxStaticText(this, wxID_STATIC, str),
               0, wxALIGN_CENTER_VERTICAL);
        s->AddStretchSpacer();
        actions[i] = new wxTextCtrl(this, wid_action0 + i);
        s->Add(actions[i], 0, wxALIGN_RIGHT);
        sizer->Add(s, 0, wxEXPAND);
    }
    action_vals[0] = ((value & 0xff0000) >> 8) | (value & 0xff);
    action_vals[1] = ((value & 0xffff00) >> 8);
    SyncActions();

    wxSizer *s = CreateSeparatedButtonSizer(wxOK|wxCANCEL);
    sizer->Add(s, 0, wxEXPAND);

    sizer->SetSizeHints(this);
    SetSizer(sizer);
    sizer->Fit(this);
}

RRegDialog::~RRegDialog()
{
    the_master->RRegDlgDisappears(regn);
}

wxString RRegDialog::MakeCaption(int id, int n)
{
    wxString s;
    s.Printf(wxT("UNC: R%d @ %04x"), n, id);
    return s;
}

void RRegDialog::SyncAction(int i)
{
    wxString s;
    s.Printf(wxT("%04x"), action_vals[i]);
    actions[i]->ChangeValue(s);
}

void RRegDialog::SyncActions()
{
    for(int i = 0; i < 2; i++) SyncAction(i);
}

void RRegDialog::SyncActionHighByte(int from, int to)
{
    long n;
    if(!actions[from]->GetValue().ToLong(&n, 16) || n < 0 || n > 0xffff) {
        int in_pt = actions[from]->GetInsertionPoint();
        SyncAction(from);
        actions[from]->SetInsertionPoint(in_pt);
        return;
    }
    if(n == action_vals[from]) return;
    action_vals[from] = n;
    action_vals[to] = (n & 0xff00) | (action_vals[to] & 0xff);
    SyncAction(to);
}

void RRegDialog::OnOk(wxCommandEvent& event)
{
    value = 0;
    for(int i = 0; i < bits_count; i++) {
        if(bits[i]->GetValue())
            value |= reaction_reg[i].mask;
    }
    value |= (action_vals[1] << 8);
    value |= (action_vals[0] & 0xff);
    UncDeviceInterface *dev = the_collection->Find(id);
    if(dev) {
        dev->SetReactionReg(regn, value);
        the_master->Sync();
    }
    delete this;
}

void RRegDialog::OnCancel(wxCommandEvent& event)
{
    delete this;
}

void RRegDialog::OnText(wxCommandEvent& event)
{
    switch(event.GetId()) {
        case wid_action0:
            SyncActionHighByte(0, 1);
            break;
        case wid_action1:
            SyncActionHighByte(1, 0);
            break;
        default:
            event.Skip();
    }
}


BEGIN_EVENT_TABLE(RRegDialog, wxDialog)
    EVT_BUTTON(wxID_OK,     RRegDialog::OnOk)
    EVT_BUTTON(wxID_CANCEL, RRegDialog::OnCancel)
    EVT_TEXT(wxID_ANY,      RRegDialog::OnText)
END_EVENT_TABLE()

///////////////////////////////////////////////////////////////////////////
//


HwConfDialog::HwConfDialog(int a_id, UncDeviceCollection *a_coll)
    : wxDialog(NULL, -1, MakeCaption(a_id), wxDefaultPosition,
               wxDefaultSize,
               wxCAPTION|wxTHICK_FRAME|wxMINIMIZE_BOX|wxSYSTEM_MENU|
               wxCLOSE_BOX|wxDIALOG_NO_PARENT),
      id(a_id), the_collection(a_coll), conf(GetDeviceConfig(a_id)), sizer(0)
{
    int i;
    sizer = new wxBoxSizer(wxHORIZONTAL);

    //////////////////////////////////
    wxStaticBoxSizer *ur_box =
        new wxStaticBoxSizer(wxVERTICAL, this, wxT("User configuration"));
    bits_count = make_bits_checkboxes(this, ur_box, userconf_reg, &bits);
    ur_box->AddSpacer(5);
    wxBoxSizer *s2 = new wxBoxSizer(wxHORIZONTAL);
    s2->Add(new wxStaticText(this, wxID_STATIC, wxT("Input line delay:")),
            0, wxALIGN_BOTTOM);
    input_delay = new wxTextCtrl(this, wxID_NONE);
    s2->Add(input_delay);
    ur_box->Add(s2);
    ur_box->AddSpacer(5);
    ur_box->Add(new wxButton(this, wid_submit_userreg, wxT("Save")),
                             0, wxALIGN_RIGHT);
    sizer->Add(ur_box, 0, wxEXPAND);

    //////////////////////////////////
    wxStaticBoxSizer *react_box =
        new wxStaticBoxSizer(wxVERTICAL, this, wxT("Reaction registers"));
    for(i = 0; i < 8; i++) {
        wxBoxSizer *s2 = new wxBoxSizer(wxHORIZONTAL);
        react_regs[i] = new wxStaticText(this, wxID_STATIC, wxT(""));
        react_buttons[i] =
            new wxButton(this, wid_rreg_button+i, wxT(" >>> "),
                         wxDefaultPosition, wxDefaultSize, wxBU_EXACTFIT);
        s2->Add(react_regs[i], 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL);
        s2->AddStretchSpacer();
        s2->Add(react_buttons[i], 0, wxALIGN_RIGHT);
        react_box->Add(s2, 0, wxEXPAND);

        rreg_dialogs[i] = 0;
    }

    sizer->Add(react_box, 0, wxEXPAND);

    //////////////////////////////////
    wxBoxSizer *box3 = new wxBoxSizer(wxVERTICAL);

    wxStaticBoxSizer *id_box =
        new wxStaticBoxSizer(wxHORIZONTAL, this, wxT("Id"));
    
    wxString str_id;
    str_id.Printf(wxT("%04x"), id);
    dev_id = new wxTextCtrl(this, wxID_NONE, str_id);
    id_box->Add(dev_id, 0, wxALIGN_CENTER_VERTICAL);
    id_box->AddStretchSpacer();
    id_box->Add(new wxButton(this, wid_change_id, wxT("Set")));
    box3->Add(id_box);

    wxStaticBoxSizer *hardware_box =
        new wxStaticBoxSizer(wxVERTICAL, this, wxT("Hardware config"));
    wxString slot_choices[] = { wxT("None"), wxT("Input"), wxT("Output") };
    for(int i = 0; i<=1; i++) {
        s2 = new wxBoxSizer(wxHORIZONTAL);
        wxString ss;
        ss.Printf(wxT("Slot %d: "), i);
        s2->Add(new wxStaticText(this, wxID_STATIC, ss),
                0, wxALIGN_CENTER_VERTICAL);
        s2->AddStretchSpacer();
        modules[i] = new wxChoice(this, wxID_ANY,
                     wxDefaultPosition, wxDefaultSize, 3, slot_choices);
        s2->Add(modules[i]);
        hardware_box->Add(s2, 0, wxEXPAND); 
    }
    s2 = new wxBoxSizer(wxHORIZONTAL);
    s2->Add(new wxStaticText(this, wxID_STATIC, wxT("Tact length: ")),
            0, wxALIGN_CENTER_VERTICAL);
    s2->AddStretchSpacer();
    s2->Add(tact_length = new wxTextCtrl(this, wxID_NONE));
    hardware_box->Add(s2, 0, wxEXPAND); 
    hardware_box->AddSpacer(5); 
    hardware_box->Add(new wxButton(this, wid_set_hwconf, wxT("Set!")),
                      0, wxALIGN_RIGHT);
    box3->Add(hardware_box, 0, wxEXPAND);


    box3->Add(new wxButton(this, wid_save_to_eeprom, wxT("Save to EEPROM")),
              0, wxEXPAND);
    box3->Add(new wxButton(this, wid_restore_eeprom,
                           wxT("Restore from EEPROM")),
              0, wxEXPAND);
    box3->Add(new wxButton(this, wid_restore_factory,
                           wxT("Restore factory defaults")),
              0, wxEXPAND);
    box3->AddSpacer(5);
    box3->Add(new wxButton(this, wxID_EXIT, wxT("Close")), 0,
              wxALIGN_RIGHT|wxALIGN_BOTTOM);
    sizer->Add(box3);
    ///////////////////////////////////
    sizer->AddSpacer(5);

    Sync();

    sizer->SetSizeHints(this);
    SetSizer(sizer);
    sizer->Fit(this);

}

HwConfDialog::~HwConfDialog()
{
    for(int i=0; i<8; i++)
        if(rreg_dialogs[i])
            delete rreg_dialogs[i];
}

wxString HwConfDialog::MakeCaption(int id)
{
    wxString s;
    s.Printf(wxT("UNC: regs for %04x"), id);
    return s;
}

void HwConfDialog::RRegDlgDisappears(int n)
{
    rreg_dialogs[n] = 0;
}

void HwConfDialog::Sync()
{
    int i;
    UncDeviceInterface *dev = the_collection->Find(id);
    if(!dev) return;
    if(!dev->GetRegisters(&hw, &user, reacts)) return;
    for(i = 0; i < bits_count; i++) {
        bits[i]->SetValue((user & userconf_reg[i].mask) != 0);
    }
    wxString s;
    s.Printf(wxT("%02lu"), user & 0xff);
    input_delay->SetValue(s);

    for(i = 0; i < 8; i++) {
        wxString s;
        s.Printf(wxT("R%d: %08lx"), i, reacts[i]);
        react_regs[i]->SetLabel(s);
    }

    for(i=0; i<2; i++) {
        int d;
        switch((hw >> (i*4)) & 0x0f) {
            case lnt_none:   d = 0; break;
            case lnt_input:  d = 1; break;
            case lnt_output: d = 2; break;
        }
        modules[i]->SetSelection(d);
    }
    s.Printf(wxT("%ld"), (hw >> 16) & 0xffff);
    tact_length->SetValue(s);
}

void HwConfDialog::OnSubmitUserreg(wxCommandEvent& event)
{
    unsigned long user = 0;
    for(int i = 0; i < bits_count; i++) {
        if(bits[i]->GetValue())
            user |= userconf_reg[i].mask;
    }
    wxString dly = input_delay->GetValue();
    unsigned long d;
    if(!dly.ToULong(&d, 10)) d = 1;
    user |= (d & 0xff);
    UncDeviceInterface *dev = the_collection->Find(id);
    if(dev)
        dev->SetUserReg(user);
}

void HwConfDialog::OnSetId(wxCommandEvent& event)
{
    wxString stri = dev_id->GetValue();
    unsigned long d;
    if(!stri.ToULong(&d, 16)) {
        wxMessageBox(wxT("Invalid ID"), wxT("Sorry..."), wxOK|wxICON_ERROR);
        return;
    }
    UncDeviceInterface *dev = the_collection->Find(id);
    if(dev) {
        dev->SetId(d);
        id = d;
    }
}

void HwConfDialog::OnSetHardware(wxCommandEvent& event)
{
    wxString really =
        wxGetTextFromUser(wxT("Are you absolutely sure\n"
                              "you know what you're doing?\n"
                              "Type the word ``really'' to confirm"),
                          wxT("DANGEROUS! please confirm"), wxT(""), this);
    if(really != wxT("really")) {
        wxMessageBox(wxT("Not confirmed"), wxT("Cancelled"));
    } else {
        static const int modv[3] = { 0x0, 0x4, 0xf };
        unsigned long reg;
        wxString tl = tact_length->GetValue();
        if(!tl.ToULong(&reg, 10)) reg = 0x4600;
        reg <<= 16;
        reg |= modv[modules[0]->GetSelection()];
        reg |= (modv[modules[1]->GetSelection()] << 4);
        UncDeviceInterface *dev = the_collection->Find(id);
        if(dev)
            dev->SetHardwareConf(reg);
    }
}

void HwConfDialog::OnSaveToEeprom(wxCommandEvent& event)
{
    UncDeviceInterface *dev = the_collection->Find(id);
    if(dev)
       dev->SaveToEeprom();
}

void HwConfDialog::OnRestoreEeprom(wxCommandEvent& event)
{
    UncDeviceInterface *dev = the_collection->Find(id);
    if(dev) {
        dev->RestoreFromEeprom();
        Sync();
    }
}

void HwConfDialog::OnRestoreFactory(wxCommandEvent& event)
{
    UncDeviceInterface *dev = the_collection->Find(id);
    if(dev) {
        dev->RestoreFactoryDefaults();
        Sync();
    }
}

void HwConfDialog::OnButton(wxCommandEvent& event)
{
    int eid = event.GetId();
    if(eid >= wid_rreg_button && eid < (wid_rreg_button + series_wid)) {
        // this is an R reg button event
        int n = eid - wid_rreg_button;
        if(!rreg_dialogs[n])
            rreg_dialogs[n] =
                new RRegDialog(id, n, reacts[n], the_collection, this);
        rreg_dialogs[n]->Show();
        rreg_dialogs[n]->Raise();
    } else {
        event.Skip();
    }
}

void HwConfDialog::OnQuit(wxCommandEvent& event)
{
    delete this;
}


BEGIN_EVENT_TABLE(HwConfDialog, wxDialog)
    EVT_BUTTON(wid_submit_userreg,  HwConfDialog::OnSubmitUserreg)
    EVT_BUTTON(wid_change_id,       HwConfDialog::OnSetId)
    EVT_BUTTON(wid_set_hwconf,      HwConfDialog::OnSetHardware)
    EVT_BUTTON(wid_save_to_eeprom,  HwConfDialog::OnSaveToEeprom)
    EVT_BUTTON(wid_restore_eeprom,  HwConfDialog::OnRestoreEeprom)
    EVT_BUTTON(wid_restore_factory, HwConfDialog::OnRestoreFactory)
    EVT_BUTTON(wxID_EXIT,           HwConfDialog::OnQuit)
    EVT_BUTTON(wxID_ANY,            HwConfDialog::OnButton)
END_EVENT_TABLE()

///////////////////////////////////////////////////////////////////////
//

void LaunchHardwareConfigurationDialog(int id, UncDeviceCollection *devcl)
{
    HwConfDialog *dlg = new HwConfDialog(id, devcl);
    dlg->Show();
}
